/*-----------------------------------------------------------------------------

Copyright (C) 2004.

A. Ronald Gallant
Post Office Box 659
Chapel Hill NC 27514-0659
USA   

Permission to use, copy, modify, and distribute this software and its docu-
mentation for any purpose and without fee is hereby granted, provided that 
the above copyright notice appears in all copies and that both that copyright 
notice and this permission notice appear in supporting documentation.  This 
software is provided "as is" without any expressed or implied warranty.

-----------------------------------------------------------------------------*/

#define COMPILER_HAS_BOOLALPHA

#include <sstream>
#include "libsnp.h"
#include "libsmm.h"
#include "snpusr.h"
#include "snp.h"
#include "emm.h"

using namespace liberr;
using namespace libcpp;
using namespace libsnp;
using namespace libsmm;
using namespace snp;
using namespace emm;
using namespace std;


namespace {
  
  class keyword {
  private:
    string kw;
  public:
    keyword() : kw(string()) {}
    keyword(const char* str) : kw(str) {}
    string set_keyword(const char* str) {return kw = str;}
    bool operator()(const string& s) const {return s.find(kw) != string::npos;}
  };
  
  char last_char(string s) 
  {
    char c = ' ';
    for (string::iterator i=s.begin(); i!=s.end(); ++i) {
      if (!isspace(*i)) c=*i;
    }
    return c;
  }

  pair<string,string> find_ss(string s)
  {
    string zero = "0"; 
    string one = "1";
    char val = 'x'; 
    char* end = &val;
    char** endptr=&end;
    int lim = s.size();
    int b=0;
    while(b<lim && isspace(s[b])) ++b; 
    if (b==lim) {
      warn("Warning, emmparms, cannot interpret this line: \"" +s+ "\"");
      warn("Warning, emmparms, 0 1 substituted for \"" +s+ "\"");
      return pair<string,string>(zero,one);
    }
    int l=0;
    while(b+l<lim && !isspace(s[b+l])) ++l;
    string first=s.substr(b,l);
    b=b+l;
    while(b<lim && isspace(s[b])) ++b; 
    if (b==lim) { 
      warn("Warning, emmparms, cannot interpret this line: \"" +s+ "\"");
      warn("Warning, emmparms, " +first+ " 1 substituted for \"" +s+ "\"");
      strtod(first.c_str(),endptr);
      if (**endptr) error("Error, emmparms, cannot read this as numeric: "+s);
      return pair<string,string>(first,one);
    }
    l=0;
    while(b+l<lim && !isspace(s[b+l])) ++l;
    string second=s.substr(b,l);
    strtod(first.c_str(),endptr);
    if (**endptr) error("Error, emmparms, cannot read line as numeric: "+s);
    strtod(second.c_str(),endptr);
    if (**endptr) error("Error, emmparms, cannot read line as numeric: "+s);
    return pair<string,string>(first,second);
  }

  bool check_estparms(estparms ep) 
  {
    if (ep.pname.size() != 12) return false;
    if (ep.emmver != 2.0) return false;
    if (ep.objtype < 0 || ep.objtype > 3) return false;
    if (ep.proptype < 0 || ep.proptype > 1) return false;
    if (ep.lsim < 0) return false;
    if (ep.nfile < 0) return false;
    if (ep.lrho < 0) return false;
    if (ep.incfac <= 0) return false;
    const REAL x = 5.0;
    if ((x*ep.incfac)/ep.incfac != x) {
      warn("Warning, check_estparms, incfac not a (fractional) power of two");
    }
    if (ep.sclfac <= 0) return false;
    if (ep.temperature <= 0) return false;
    return true;
  }
  
}

emm::emmparms::emmparms()
: history(), epm(), rho(), rho_mask(), rho_increment(), prop_scale(),
  prop_groups(), usrparms(), rv(false)
{ }

void emm::emmparms::set_estparms(const estparms& e)
{  
  if (check_estparms(e)) {
    epm = e;
  }
  else {
    error("Error, emmparms, set_estparms, bad input");
  }
}

void emm::emmparms::set_rho(const realmat& r)
{  
  if (rho.get_rows() != r.get_rows() || rho.get_cols() != r.get_cols() ) {
    rho = r;
  }
  else {
    error("Error, emmparms, set_rho, bad input");
  }
}

bool emm::emmparms::read_parms(const char* filename, ostream& detail)
{
  history.erase(history.begin(),history.end());
  usrparms.erase(usrparms.begin(),usrparms.end());
  grouptxt.erase(grouptxt.begin(),grouptxt.end());
  parmrhs.erase(parmrhs.begin(),parmrhs.end());
  scalerhs.erase(scalerhs.begin(),scalerhs.end());
  incrhs.erase(incrhs.begin(),incrhs.end());
  
  ifstream pfstream(filename);
  if (!pfstream) return rv = false;

  vector<string> pf; 
  typedef vector<string>::const_iterator pf_itr;

  string line;
  while (getline(pfstream, line)) pf.push_back(line);

  keyword kw;
  string header;

  pf_itr kw_ptr;
  
  header=kw.set_keyword("PARMFILE HISTORY");  //optional
  kw_ptr = find_if(pf.begin(), pf.end(), kw);
  if (kw_ptr != pf.end()) {
    kw_ptr += 8;    // Number of lines output history
    if (kw_ptr > pf.end()) error("Error, emmparms, " + header);
    while (kw_ptr != pf.end() && (*kw_ptr)[0] == '#') {
      history.push_back(*kw_ptr); 
      ++kw_ptr;
    }
  }
  
  header=kw.set_keyword("ESTIMATION DESCRIPTION");  //required
  kw_ptr = find_if(pf.begin(), pf.end(), kw);
  if (kw_ptr == pf.end()) return rv = false;
  if (kw_ptr + 14 >= pf.end()) error("Error, emmparms, " + header);

  epm.pname = (++kw_ptr)->substr(0,12); 
  epm.emmver = atof((++kw_ptr)->substr(0,12).c_str());
  string ss = *(++kw_ptr);
  epm.score = cutstr(ss);;
  epm.objtype = atoi((++kw_ptr)->substr(0,12).c_str()); 
  epm.proptype = atoi((++kw_ptr)->substr(0,12).c_str()); 
  epm.print = atoi((++kw_ptr)->substr(0,12).c_str()); 
  epm.seed = atoi((++kw_ptr)->substr(0,12).c_str()); 
  epm.lsim = atoi((++kw_ptr)->substr(0,12).c_str()); 
  epm.nfile = atoi((++kw_ptr)->substr(0,12).c_str()); 
  epm.lrho = atoi((++kw_ptr)->substr(0,12).c_str());
  epm.incfac = atof((++kw_ptr)->substr(0,12).c_str());
  epm.sclfac = atof((++kw_ptr)->substr(0,12).c_str());
  epm.temperature = atof((++kw_ptr)->substr(0,12).c_str());
  epm.kilse = atoi((++kw_ptr)->substr(0,12).c_str());

  if (epm.print) {
    detail << starbox("/Input EMM parmfile//") << '\n';
    for (pf_itr itr=pf.begin(); itr!=pf.end(); ++itr) {
      detail << *itr << '\n';
    }
    detail.flush();
    detail << starbox("/EMM parmfile interpretation//") << '\n';
  }

  if (epm.print) {
    detail << '\n' << header << '\n';
    #if defined COMPILER_HAS_BOOLALPHA
      detail << boolalpha;
    #endif
    detail << "\t pname = " << epm.pname << '\n';
    detail << "\t emmver = " << fmt('f',3,1,epm.emmver) << '\n';
    detail << "\t score = " << epm.score << '\n';
    detail << "\t objtype = " << epm.objtype << '\n';
    detail << "\t proptype = " << epm.proptype << '\n';
    detail << "\t print = " << epm.print << '\n';
    detail << "\t seed = " << epm.seed << '\n';
    detail << "\t lsim = " << epm.lsim << '\n';
    detail << "\t nfile = " << epm.nfile << '\n';
    detail << "\t lrho = " << epm.lrho << '\n';
    detail << "\t incfac = " << epm.incfac 
           << " (should have exact binary representation)\n";
    detail << "\t sclfac = " << epm.sclfac << '\n';
    detail << "\t temperature = " << epm.temperature << '\n';
    detail << "\t kilse = " << epm.kilse << '\n';
    detail.flush();
  }

  if (!check_estparms(epm)) {
     warn(string("Warning, bad or too few items in ")+header);
     return rv = false;
  }

  if (epm.print) {
    detail << starbox("/Attempt to read SNP parmfile/SNP now controls print//");
    detail.flush();
  }
  
  parmfile snp_pf;

  if(!snp_pf.read_parms(epm.score.c_str(),detail)) {
     detail.flush();
     error("Error, emmparms, cannot read snp parmfile");
  }

  datread_type dr; 
  dr.initialize(snp_pf.get_datparms());

  if (!dr.read_data(data)) error("Error, emmparms, read_data failed");

  if (epm.print) {
    detail << starbox("/Test read of score succeeded/EMM now controls print//");
    detail << starbox("/First 12 observations//");
    detail << data("",seq(1,12));
    detail << starbox("/Last 12 observations//");
    detail << data("",seq(snp_pf.get_datparms().n-11, snp_pf.get_datparms().n));
    detail.flush();
  }

  rho.resize(epm.lrho,1);
  rho_mask.resize(rho.size());
  
  INTEGER nfixed = 0;

  header=kw.set_keyword("PARAMETER START VALUES");   //required
  kw_ptr = find_if(pf.begin(), pf.end(), kw);
  if (kw_ptr + rho.size() >= pf.end()) error("Error, emmparms, " + header);
  for (INTEGER i=1; i<=rho.size(); ++i) {
    pair<string,string> ss = find_ss((++kw_ptr)->c_str());
    rho[i] = atof(ss.first.c_str());
    rho_mask[i] = atoi(ss.second.c_str());
    if (rho_mask[i] == 0) ++nfixed;
    string line = *kw_ptr;
    cutstr(line); cutstr(line);
    parmrhs.push_back(line);
  }
  
  INTEGER nactive = rho.size() - nfixed;

  intvec fixed;
  intvec active = seq(1,rho.size());
  if (nfixed>0) {
    fixed.resize(nfixed);
    active.resize(nactive);
    INTEGER j=0;
    INTEGER k=0;
    for (INTEGER i=1; i<=rho.size(); ++i) {
      if (rho_mask[i] == 0) {
        fixed[++j]=i;
      }
      else {
        active[++k]=i;
      }
    }
  }

  if (epm.print) {
    detail << '\n' << header << '\n';
    detail << "\n\t rho = ";
    detail << rho;
    detail << "\n\t rho_mask = ";
    detail << rho_mask << '\n';
    detail << "\t Number fixed = " << nfixed << '\n';
    detail << "\t Number active = " << nactive << '\n';
    detail << "\n\t fixed = ";
    detail << fixed << '\n';
    detail << "\n\t active = ";
    detail << active << '\n';
    detail.flush();
  }

  rho_increment.resize(rho.size(),1);
  
  header=kw.set_keyword("PARAMETER INCREMENTS");   //required
  kw_ptr = find_if(pf.begin(), pf.end(), kw);
  if (kw_ptr + rho.size() >= pf.end()) error("Error, emmparms, " + header);
  for (INTEGER i=1; i<=rho.size(); ++i) {
    rho_increment[i] = atof((++kw_ptr)->c_str());
    rho_increment[i] *= epm.incfac;
    if (rho_increment[i]<=0) error("Error, emmparms, " + header);
    string line = *kw_ptr;
    cutstr(line);
    incrhs.push_back(line);
  }

  if (epm.print) {
    detail << '\n' << header << '\n';
    detail << "(after multiplication by incfac)\n";
    detail << "(output parmfile this increment and incfac=1.0)\n";
    detail << "(should have exact binary representation)\n";
    detail << "\n\t rho_increment = ";
    detail << rho_increment;
    detail.flush();
  }

  prop_scale.resize(rho.size(),1);
  
  header=kw.set_keyword("PROPOSAL SCALING");   //required
  kw_ptr = find_if(pf.begin(), pf.end(), kw);
  if (kw_ptr + rho.size() >= pf.end()) error("Error, emmparms, " + header);
  for (INTEGER i=1; i<=rho.size(); ++i) {
    prop_scale[i] = atof((++kw_ptr)->c_str());
    prop_scale[i] *= epm.sclfac;
    if (prop_scale[i]<=0) error("Error, emmparms, " + header);
    string line = *kw_ptr;
    cutstr(line);
    scalerhs.push_back(line);
  }

  if (epm.print) {
    detail << '\n' << header << '\n';
    detail << "(after multiplication by sclfac)\n";
    detail << "(output parmfile this scale and sclfac=1.0)\n";
    detail << "\n\t prop_scale = ";
    detail << prop_scale;
    detail.flush();
  }

  header=kw.set_keyword("USER PROBLEM DESCRIPTION");  //optional
  kw_ptr = find_if(pf.begin(), pf.end(), kw);
  if (kw_ptr != pf.end()) {
    ++kw_ptr;
    while (kw_ptr != pf.end()) {
      usrparms.push_back(*kw_ptr); 
      ++kw_ptr;
    }
  }

  if (epm.print) {
    detail << '\n' << header << '\n';
    vector<std::string>::const_iterator itr;
    for (itr=usrparms.begin(); itr != usrparms.end(); ++itr) {
      detail << *itr << '\n';
    }
    detail.flush();
  }

  if (nfixed > 0) {
    realmat ufixed(nfixed,1);
    realmat Vfixed(nfixed,nfixed,0.0);
    for (INTEGER i=1; i<=nfixed; ++i) {
      ufixed[i] = rho[fixed[i]];
      Vfixed(i,i) = pow(prop_scale[fixed[i]],2);
    }
    prop_groups.push_back(prop_group(0.0,fixed,ufixed,Vfixed));
  }

  header=kw.set_keyword("PROPOSAL GROUPING");  //optional
  kw_ptr = find_if(pf.begin(), pf.end(), kw);
  if (kw_ptr != pf.end()) {
    INTEGER count = 0;
    while (count < rho.size()) {
      INTEGER top = rho.size();
      INTEGER r=1;
      realmat gmat(rho.size()+1,rho.size()+1,0.0);
      while (r <= top) {
        ++kw_ptr;
        grouptxt.push_back(*kw_ptr);
        if (kw_ptr == pf.end()) error("Error, emmparms, " + header); 
        INTEGER j=0;
        INTEGER c=0;
        INTEGER lim = kw_ptr->size();
	while(lim > 0 && isspace((*kw_ptr)[lim-1])) --lim;
        while (j < lim && c <= rho.size()) {
          while (j < lim && isspace((*kw_ptr)[j])) ++j;
          INTEGER k=0;
          while (j+k < lim && !isspace((*kw_ptr)[j+k])) ++k;
          gmat(r,++c) = atof(kw_ptr->substr(j,k).c_str());
          j+=k;
        }
        if (r != 1) ++count;
        if (r == 1) top = c;
        ++r;
      }
      intvec idx = seq(1,top);
      gmat = gmat(idx,idx);
      for (INTEGER i=1; i<=top; ++i) {
        if (gmat(1,i) != gmat(i,1)) error("Error, emmparms, " + header);
      }
      for (INTEGER i=2; i<=top; ++i) {
        for (INTEGER j=1; j<=fixed.size(); ++j) {
          if (INTEGER(gmat(1,idx[i])) == fixed[j]) idx[i] = -idx[i];
        }
      }
      realmat gmat_active = gmat(idx,idx);
      if (gmat_active.size() > 1) {
        idx[1] = -idx[1];
        REAL freq = gmat(1,1);
        realmat rvec = gmat(1,idx);
        realmat Vmat = gmat(idx,idx);
        intvec gvec(rvec.size());
        realmat mean(rvec.size(),1);
        for (INTEGER i=1; i<=gvec.size(); ++i) {
          gvec[i] = INTEGER(rvec[i]);
          mean[i] = rho[gvec[i]];
        }
        for (INTEGER j=1; j<=gvec.size(); ++j) {
          for (INTEGER i=1; i<=gvec.size(); ++i) {
            Vmat(i,j) = Vmat(i,j)*prop_scale[gvec[i]]*prop_scale[gvec[j]];
          }
        }
        prop_groups.push_back(prop_group(freq, gvec, mean, Vmat));
      }
    }
  }
  else {
    for (INTEGER i=1; i<=nactive; ++i) {
      intvec gvec(1,active[i]);
      REAL var = pow(prop_scale[active[i]],2);
      REAL mu = rho[active[i]];
      prop_groups.push_back
        ( prop_group( 1.0, gvec,realmat(1,1,mu), realmat(1,1,var) ) );
    }
  }

  if (epm.print) {
    detail << starbox("/Prop groups//") << '\n';
    detail << "\t\t Constructed from "<<header<<", if present; \n";
    detail << "\t\t If not, move one at a time grouping is used.\n";
    detail << "\t\t If rho has fixed elements, they are in Group 0\n";
    detail << "\t\t and not changed during the MCMC iterations.\n";
    detail << "\t\t "<<header<<" correlations have been converted\n";
    detail << "\t\t to variances using the prop_scale vector above.\n";
    detail << "\t\t Frequencies are relative, not absolute.\n\n";
    prop_def::const_iterator itr;
    INTEGER i=0;
    for (itr=prop_groups.begin(); itr!=prop_groups.end(); ++itr) {
      detail << "     Group " << i++ << "\n\n"; 
      detail << "\t freq = " << itr->freq << '\n';
      detail << "\t gvec = " << itr->gvec << '\n';
      detail << "\t mean = " << itr->mean << '\n';
      detail << "\t Vmat = " << itr->Vmat << '\n';
    }
  }

  return rv = true;
}

bool emm::emmparms::write_parms (string parmfile, string prefix,
                 const realmat& mu, const realmat& sig) const
{
  string filename = prefix + ".parmfile.alt";
  ofstream ps(filename.c_str());
  if (!ps) return false; 

  ps << "PARMFILE HISTORY (optional)\n"
     << "#\n"
     << "# This parmfile was written by EMM Version " 
     << fmt('f',3,1,epm.emmver) << ' '
     << "using the following line from\n" 
     << "# control.dat, which was read as " 
     << "char*, char*\n"
     << "# -----------------------------------------"
     << "----------------------------------\n";
//ps << "#      input_parmfile       output_prefix\n"   // Change kw_ptr
//   << "#      --------------       -------------\n";  // line if altered
  ps << '#';
  for (INTEGER i=1; i<=INTEGER(20-parmfile.size()); ++i) ps << ' ';
  ps << parmfile;
  for (INTEGER i=1; i<=INTEGER(20-prefix.size()); ++i) ps << ' ';
  ps << prefix << '\n';
  ps << "# -----------------------------------------"
     << "----------------------------------\n"
     << "#\n";
  for (vector<string>::const_iterator i=history.begin();i!=history.end();++i) {
    ps << *i << '\n';
  }

  ps << "ESTIMATION DESCRIPTION (required)\n"
     << epm.pname << "   "
     << "Project name, pname, char*\n"
     << fmt('f', 12, 1, epm.emmver) <<  "   "
     << "EMM version, defines format of this file, emmver, float\n";
  for (INTEGER i=1; i<=12-INTEGER(epm.score.size()); ++i) ps << ' '; 
  ps << epm.score  <<  "   "
     << "SNP parmfile, initializes EMM, EDF, score, char*\n"
     << fmt('d', 12, epm.objtype) <<  "   "
     << "Objfun type, 0 EMM, 1 EDF, 2 usr, objtype, int\n"
     << fmt('d', 12, 1) <<  "   "
     << "Proposal type, 0 group_move, 1 cond_move, 2 usr, proptype, int\n"
     << fmt('d', 12, epm.print) <<  "   "
     << "Write detailed output if print=1, int\n"
     << fmt('d', 12, epm.seed) <<  "   "
     << "Seed for simulations, iseed, int\n"
     << fmt('d', 12, epm.lsim) <<  "   " 
     << "Number MCMC simulations per file, lsim, int\n"
     << fmt('d', 12, epm.nfile) <<  "   "
     << "Number of MCMC simulation files beyond the first, nfile, int\n"
     << fmt('d', 12, epm.lrho) <<  "   "
     << "Number of parameters, lrho, int\n"
     << fmt('f', 12, 1, 1.0) << "   "
     << "Rescale parameter increments by this value, incfac, float\n"
     << fmt('f', 12, 1, 1.0) << "   "
     << "Rescale proposal scaling by this value, sclfac, float\n";
  if (epm.temperature == 1.0) {
    ps << fmt('f', 12, 1, epm.temperature) << "   ";
  } 
  else {
    ps << fmt('g', 12, 5, epm.temperature) << "   ";
  }
  ps << "Rescale objfun by this value, temperature, float\n"
     << fmt('d', 12, epm.kilse) << "   "
     << "Statistics not computed if kilse=1, int\n";

  realmat corr(epm.lrho,epm.lrho,0.0);
  realmat rho_scale(epm.lrho,1);
  for (INTEGER i=1; i<=epm.lrho; ++i) {
    rho_scale[i] = sqrt(sig(i,i));
    for (INTEGER j=1; j<=epm.lrho; ++j) {
      if (sig(i,i) > 0 && sig(j,j) > 0) {
        corr(i,j) = sig(i,j)/sqrt(sig(i,i)*sig(j,j));
      }
    }
  }

  vector<string>::const_iterator itr;

  ps << "PARAMETER START VALUES (required)\n";
  itr = parmrhs.begin();
  for (INTEGER i=1; i<=epm.lrho; ++i) {
    ps << fmt('e',26,17,mu[i]) << fmt('d',5, rho_mask[i]) 
       << (itr++)->substr(0,49) << '\n';
  }

  ps << "PARAMETER INCREMENTS (required) " 
     << "(fractional powers of two recommended)\n";
  itr = incrhs.begin();
  for (INTEGER i=1; i<=epm.lrho; ++i) {
    ps << fmt('e',26,17,rho_increment[i]) << (itr++)->substr(0,49) << '\n';
  }

  string stop = "PROPOSAL SCALING (required)";
  ps << stop << '\n';
  itr = scalerhs.begin();
  for (INTEGER i=1; i<=epm.lrho; ++i) {
    ps << fmt('e',26,17,rho_scale[i]) << (itr++)->substr(0,49) << '\n';
  }

  ps << "PROPOSAL GROUPING (optional) (frequencies are relative)\n";
  ps << " 1.0 ";
  for (INTEGER j=1; j<=epm.lrho; ++j) ps << j << " ";
  ps << '\n';
  for (INTEGER i=1; i<=epm.lrho; ++i) {
    ps << "  " << i << ' ';
    for (INTEGER j=1; j<=epm.lrho; ++j) {
      ps << fmt('e',26,17,corr(i,j)) << ' ';  
    }
    ps << '\n'; 
  }

  ps << "USER PROBLEM DESCRIPTION "
     << "(optional, must be last block of parmfile)\n";
  vector<string>::const_iterator usritr;
  for (usritr=usrparms.begin(); usritr!=usrparms.end(); ++usritr){
    ps << *usritr << '\n';
  }

  ps.close();

  ifstream qs(filename.c_str());
  if (!qs) return false; 

  filename = prefix + ".parmfile.fit";
  ofstream rs(filename.c_str());
  if (!rs) return false; 

  string line;
  while(getline(qs,line) && line != stop) {
    if (line.find("Proposal type") != string::npos) {
      rs << fmt('d', 12, epm.proptype) <<  "   "
         << "Proposal type, 0 group_move, 1 cond_move, 2 usr, proptype, int\n";
    }
    else {
      rs << line << '\n';
    }
  }

  rs << stop << '\n';
  itr = scalerhs.begin();
  for (INTEGER i=1; i<=epm.lrho; ++i) {
    rs << fmt('e',26,17,prop_scale[i]) << (itr++)->substr(0,49) << '\n';
  }

  if (grouptxt.size() > 0) {
    rs << "PROPOSAL GROUPING (optional) (frequencies are relative)\n";
    vector<string>::const_iterator grpitr;
    for (grpitr=grouptxt.begin(); grpitr!=grouptxt.end(); ++grpitr){ 
      rs << *grpitr << '\n';
    }
  }

  rs << "USER PROBLEM DESCRIPTION "
     << "(optional, must be last block of parmfile)\n";
  for (usritr=usrparms.begin(); usritr!=usrparms.end(); ++usritr){
    rs << *usritr << '\n';
  }


  return true;
}

