/*-----------------------------------------------------------------------------

Copyright (C) 2004.

A. Ronald Gallant
Post Office Box 659
Chapel Hill NC 27514-0659
USA   

Permission to use, copy, modify, and distribute this software and its docu-
mentation for any purpose and without fee is hereby granted, provided that 
the above copyright notice appears in all copies and that both that copyright 
notice and this permission notice appear in supporting documentation.  This 
software is provided "as is" without any expressed or implied warranty.

-----------------------------------------------------------------------------*/

#include "libsnp.h"
#include "libsmm.h"
#include "snpusr.h"
#include "snp.h"
#include "emm.h"

using namespace liberr;
using namespace libcpp;
using namespace libsnp;
using namespace libsmm;
using namespace snp;
using namespace emm;
using namespace std;

namespace {

  snpll nlopt_ll;

  class objective_function : public nleqns_base {
  private:
    INTEGER count;
  public:
    objective_function() : count(0) { }
    void reset_count() { count = 0; }
    INTEGER get_count() { return count; }
    bool get_f(const realmat& rho, realmat& f)
    {
      if (rows(f)!=1) f.resize(1,1);
      nlopt_ll.set_rho(rho);
      realmat dllwrho;
      REAL log_like = nlopt_ll.log_likehood(dllwrho);
      f[1] = -log_like/nlopt_ll.get_datparms().n;
      ++count;
      return true;
    }
    bool get_F(const realmat& rho, realmat& f, realmat& F)
    {
      if (rows(f)!=1) f.resize(1,1);
      nlopt_ll.set_rho(rho);
      realmat dllwrho;
      REAL log_like = nlopt_ll.log_likehood(dllwrho);
      f[1] = -log_like/nlopt_ll.get_datparms().n;
      F = -dllwrho/nlopt_ll.get_datparms().n;
      ++count;
      return true;
    }
  };
}

emm::emm_objfun::emm_objfun(const char* filename, std::ostream& detail)
{
  parmfile pf;

  if(!pf.read_parms(filename,detail)) {
     if (pf.get_optparms().print) detail.flush();
     error("Error, emm_objfun, cannot read parmfile");
  }

  string msg = "Warning, emm_objfun, " + string(filename);

  iter = pf.get_optparms().itmax0; 
  if (iter < 15) {
    warn(msg + string(" value for itmax0 too small, iter set to 15"));
    iter = 15;
  }

  toler = pf.get_optparms().toler; 
  if (toler < sqrt(EPS) || toler > 0.1) {
    warn(msg + string(" value for toler invalid, reset to root EPS"));
    toler = sqrt(EPS);
  }
    
  realmat Y;
  
  datread_type dr; 
  dr.initialize(pf.get_datparms());
  
  if (!dr.read_data(Y)) error("Error, emm_objfun, read_data failed");

  if (pf.get_optparms().print) {
    detail << starbox("/First 12 observations//");
    detail << Y("",seq(1,12));
    detail << starbox("/Last 12 observations//");
    detail << Y("",seq(pf.get_datparms().n-11, pf.get_datparms().n));
  }

  trnfrm tr_tmp(pf.get_datparms(), pf.get_tranparms(), Y);
  tr = tr_tmp;

  pf.set_tranparms(tr.get_tranparms());
    
  if (pf.get_optparms().print) {
    detail << starbox("/Mean and variance of data//");
    if (pf.get_tranparms().diag) {
      detail << "(Variance has been diagonalized.)\n";
    }
    detail << pf.get_tranparms().mean << pf.get_tranparms().variance;
  }

  realmat X = Y;

  tr.normalize(Y);
  tr.normalize(X);

  if (pf.get_tranparms().squash == 1) {
    tr.spline(X);
  } else if (pf.get_tranparms().squash == 2) {
    tr.logistic(X);
  }

  if (pf.get_optparms().print) {
    detail << starbox("/First 12 normalized observations//");
    detail << Y("",seq(1,12));
    detail << starbox("/Last 12 normalized observations//");
    detail << Y("",seq(pf.get_datparms().n-11, pf.get_datparms().n));
    if (pf.get_tranparms().squash > 0) {
      detail << starbox("/First 12 transformed observations//");
      detail << Y("",seq(1,12));
      detail << starbox("/Last 12 transformed observations//");
      detail << Y("",seq(pf.get_datparms().n-11, pf.get_datparms().n));
    }
    detail.flush();
  }

  snpll ll_tmp(pf.get_datparms(),
    pf.get_snpden(), pf.get_afunc(), pf.get_ufunc(),
    pf.get_rfunc(), pf.get_afunc_mask(), pf.get_ufunc_mask(),
    pf.get_rfunc_mask());

  ll = ll_tmp;
  ll.set_XY(&X,&Y);

  realmat dllwrho;
  ll.log_likehood(dllwrho, rho_infmat);
  rho_infmat = rho_infmat/ll.get_datparms().n;
  V = invpsd(rho_infmat);
}

emm::emm_objfun::emm_objfun() 
: tr(), ll(), V(), iter(), toler() 
{ error("Error, emm_objfun, default constructor got called"); }

emm::emm_objfun::emm_objfun(const emm_objfun& obj)
: tr(obj.tr), ll(obj.ll), V(obj.V), iter(obj.iter), toler(obj.toler) 
{ }

emm::emm_objfun& emm_objfun::operator=(const emm_objfun& obj)
{
  if (this != &obj) {
    tr = obj.tr; ll = obj.ll; V = obj.V; iter = obj.iter; toler = obj.toler;
  }
  return *this;
}

void emm::emm_objfun::set_parms(const realmat& data)
{
  nlopt_ll = ll;

  realmat Y = data;

  nlopt_ll.set_n(Y.get_cols());
  
  tr.normalize(Y);

  realmat X;
  
  if (tr.get_tranparms().squash == 1) {
    X = Y;
    tr.spline(X);
    nlopt_ll.set_XY(&X,&Y);
  }
  else if (tr.get_tranparms().squash == 2) {
    X = Y;
    tr.logistic(X);
    nlopt_ll.set_XY(&X,&Y);
  }
  else {
    nlopt_ll.set_XY(&Y,&Y);
  }
  
  objective_function obj;
  nlopt minimizer(obj);  
  
  minimizer.set_iter_limit(iter);
  minimizer.set_solution_tolerance(toler);

  realmat rho_start = nlopt_ll.get_rho();
  realmat rho_stop  = rho_start; 

  minimizer.minimize(rho_start, rho_stop);

  ll.set_rho(rho_stop);
  nlopt_ll.set_rho(rho_stop);

  realmat dllwrho;
  nlopt_ll.log_likehood(dllwrho, rho_infmat);
  rho_infmat = rho_infmat/nlopt_ll.get_datparms().n;

  #if defined GNU_GPP_COMPILER
    bool bad_sim = false;
    for (INTEGER i=1; i<=rho_infmat.size(); ++i) {
        if (!finite(rho_infmat[i])) bad_sim = true;
    }
    if (bad_sim) { 
      warn("Warning, emm_objfun, set_parms, bad simulation");
      V = inv(rho_infmat);
    }
    else {
      V = invpsd(rho_infmat);
    }  
  #else
    V = invpsd(rho_infmat);
  #endif
}

REAL emm::emm_objfun::operator()(const realmat& sim) const
{ 
  nlopt_ll = ll;

  realmat Y = sim;

  nlopt_ll.set_n(Y.get_cols());

  tr.normalize(Y);
  
  realmat X;
  
  if (tr.get_tranparms().squash ==1) {
    X = Y;
    tr.spline(X);
    nlopt_ll.set_XY(&X,&Y);
  }
  else if (tr.get_tranparms().squash == 2) {
    X = Y;
    tr.logistic(X);
    nlopt_ll.set_XY(&X,&Y);
  }
  else {
    nlopt_ll.set_XY(&Y,&Y);
  }
  
  realmat dllwrho;
  nlopt_ll.log_likehood(dllwrho);
  mean_score = dllwrho/Y.get_cols();
  realmat sn = (mean_score*V)*T(mean_score);

  REAL rv = sn[1];
  
  rv *= ll.get_datparms().n;

  return rv;
}  

bool emm::emm_objfun::write_diagnostics(const char* filename) const
{
  ofstream os(filename);
  if (!os) {
     std::string msg("Error, realmat, write_diagnostics, cannot open ");
     error(msg + filename);
  }

  os << "Score diagnostics:\n";

  realmat rho = ll.get_rho();
  realmat theta = ll.get_theta();

  snpden f=ll.get_snpden();
  afunc af=ll.get_afunc();
  ufunc uf=ll.get_ufunc();
  rfunc rf=ll.get_rfunc();

  REAL n = ll.get_datparms().n;

  realmat se(theta.size(),1,0.0);
  realmat tv(theta.size(),1,0.0);
  realmat sc(theta.size(),1,0.0);

  realmat V = rho_infmat;
  cerr << V.size() << '\n';

  vector< std::pair<INTEGER,INTEGER> > rt = ll.get_rt();

  vector< pair<INTEGER,INTEGER> >::const_iterator rt_ptr;
  for (rt_ptr = rt.begin(); rt_ptr != rt.end(); ++rt_ptr) {
    se[rt_ptr->second] = sqrt(V(rt_ptr->first,rt_ptr->first));
    sc[rt_ptr->second] = sqrt(n)*mean_score[rt_ptr->first];
  }
  for (INTEGER i=1; i<=theta.size(); ++i) {
    tv[i] = (se[i]==0.0 ? 0.0 : sc[i]/se[i]);
  }

  os << "           normalized        standard\n";
  os << "Index      mean score           error     t-statistic    descriptor\n";
  INTEGER count=0;
  for (INTEGER i=1; i<=af.get_nrowa0(); ++i) {
    ++count;
    os << fmt('d',5,count) << ' '
       << fmt('f',15,5,sc[count]) << ' '
       << fmt('f',15,5,se[count]) << ' '
       << fmt('f',15,5,tv[count])
       << "    a0["<<i<<"]   ";
    for(INTEGER k=1; k<=f.get_ly(); ++k) {
      os << af.get_alpha()[af.get_odx()[i]-1][k];
    }
    os << '\n';
  }
  for (INTEGER j=1; j<=af.get_ncolA(); ++j) {
    for (INTEGER i=1; i<=af.get_nrowA(); ++i) {
      ++count;
      os << fmt('d',5,count) << ' '
	 << fmt('f',15,5,sc[count]) << ' ' 
	 << fmt('f',15,5,se[count]) << ' '
	 << fmt('f',15,5,tv[count])
         << "    A("<<i<<','<<j<<")  ";
      for(INTEGER k=1; k<=f.get_ly(); ++k) {
        os << af.get_alpha()[af.get_idx()[i]-1][k];
      }
      os << " ";
      for(INTEGER k=1; k<=af.get_lx()*af.get_lag(); ++k) {
        os << af.get_beta()[j-1][k];
      }
      os << '\n';
    }
  }
  if (uf.is_intercept()) {
    for (INTEGER i=1; i<=uf.get_ly(); ++i) {
      ++count;
      os << fmt('d',5,count) << ' '
	 << fmt('f',15,5,sc[count]) << ' ' 
	 << fmt('f',15,5,se[count]) << ' '
	 << fmt('f',15,5,tv[count])
         << "    b0["<<i<<"] ";
      os << '\n';
    }
  }
  if (uf.is_regression()) {
    for (INTEGER j=1; j<=uf.get_lx()*uf.get_lag(); ++j) {
      for (INTEGER i=1; i<=uf.get_ly(); ++i) {
	++count;
        os << fmt('d',5,count) << ' '
	   << fmt('f',15,5,sc[count]) << ' ' 
	   << fmt('f',15,5,se[count]) << ' '
	   << fmt('f',15,5,tv[count])
	   << "    B("<<i<<','<<j<<") ";
	os << '\n';
      }
    }
  }
  for (INTEGER i=1; i<=rf.get_lR(); ++i) {
    ++count;
    os << fmt('d',5,count) << ' '
       << fmt('f',15,5,sc[count]) << ' ' 
       << fmt('f',15,5,se[count]) << ' '
       << fmt('f',15,5,tv[count])
       << "    R0["<<i<<"] ";
    os << '\n';
  }
  for (INTEGER j=1; j<=rf.get_colsP(); ++j) {
    for (INTEGER i=1; i<=rf.get_rowsP(); ++i) {
      ++count;
      os << fmt('d',5,count) << ' '
	 << fmt('f',15,5,sc[count]) << ' ' 
	 << fmt('f',15,5,se[count]) << ' '
	 << fmt('f',15,5,tv[count])
         << "    P("<<i<<','<<j<<")  " << rf.get_Ptype();
      os << '\n';
    }
  }
  for (INTEGER j=1; j<=rf.get_colsQ(); ++j) {
    for (INTEGER i=1; i<=rf.get_rowsQ(); ++i) {
      ++count;
      os << fmt('d',5,count) << ' '
	 << fmt('f',15,5,sc[count]) << ' ' 
	 << fmt('f',15,5,se[count]) << ' '
	 << fmt('f',15,5,tv[count])
         << "    Q("<<i<<','<<j<<")  " << rf.get_Qtype();
      os << '\n';
    }
  }
  for (INTEGER j=1; j<=rf.get_colsV(); ++j) {
    for (INTEGER i=1; i<=rf.get_rowsV(); ++i) {
      ++count;
      os << fmt('d',5,count) << ' '
	 << fmt('f',15,5,sc[count]) << ' ' 
	 << fmt('f',15,5,se[count]) << ' '
	 << fmt('f',15,5,tv[count])
         << "    V("<<i<<','<<j<<")  " << rf.get_Vtype();
      os << '\n';
    }
  }
  for (INTEGER j=1; j<=rf.get_colsW(); ++j) {
    for (INTEGER i=1; i<=rf.get_rowsW(); ++i) {
      ++count;
      os << fmt('d',5,count) << ' '
	 << fmt('f',15,5,sc[count]) << ' ' 
	 << fmt('f',15,5,se[count]) << ' '
	 << fmt('f',15,5,tv[count])
         << "    W("<<i<<','<<j<<")  " << rf.get_Wtype();
      os << '\n';
    }
  }

  os.flush();
  return os.good();
}

objfun_base* emm::emm_objfun::new_objfun()
{ 
  objfun_base* r = new(nothrow) emm_objfun(*this);
  if (r == 0) error("Error, emm_objfun, operator new failed");
  return r;
}

void emm::emm_objfun::delete_objfun(objfun_base* objfun_ptr)
{ 
  delete objfun_ptr;
}


emm::edf_objfun::edf_objfun(const char* filename, std::ostream& detail)
{
  parmfile pf;

  if(!pf.read_parms(filename,detail)) {
     if (pf.get_optparms().print) detail.flush();
     error("Error, emm_objfun, cannot read parmfile");
  }

  realmat Y;
  
  datread_type dr; 
  dr.initialize(pf.get_datparms());

  if (!dr.read_data(Y)) error("Error, emm_objfun, read_data failed");
  
  INTEGER lag_af = pf.get_afunc().get_lag();
  INTEGER lag_uf = pf.get_ufunc().get_lag();
  INTEGER lag_rf_P = pf.get_rfunc().get_p();
  INTEGER lag_rf_Q = pf.get_rfunc().get_q();
  INTEGER lag_rf_V = pf.get_rfunc().get_v();
  INTEGER lag_rf_W = pf.get_rfunc().get_w();

  lag = lag_af > lag_uf ? lag_af : lag_uf;
  lag = lag > lag_rf_P ? lag : lag_rf_P;
  lag = lag > lag_rf_Q ? lag : lag_rf_Q;
  lag = lag > lag_rf_V ? lag : lag_rf_V;
  lag = lag > lag_rf_W ? lag : lag_rf_W;

  this->set_parms(Y);
}

emm::edf_objfun::edf_objfun() 
: data(), edf_dat(), d(), lag() 
{ error("Error, edf_objfun, default constructor got called"); }

emm::edf_objfun::edf_objfun(const edf_objfun& obj)
: data(obj.data), edf_dat(obj.edf_dat), d(obj.d), lag(obj.lag)
{ }

edf_objfun& emm::edf_objfun::operator=(const edf_objfun& obj)
{
  if (this != &obj) {
    data = obj.data; edf_dat = obj.edf_dat; d = obj.d; lag = obj.lag;
  }
  return *this;
}

void emm::edf_objfun::set_parms(const realmat& dat)
{
  d = rows(dat);
  INTEGER n = cols(dat);

  data.resize((lag+1)*d,n-lag);

  for (INTEGER t=lag+1; t<=n; t++) {
    for (INTEGER j=0; j<=lag; ++j) {
      for (INTEGER i=1; i<=d; i++) {
        data(j*d+i,t-lag) = dat(i,t-j);
      }
    }
  }

  libcpp::edf(data,data,edf_dat); 
}

REAL emm::edf_objfun::operator()(const realmat& sim) const
{
  if (rows(sim) != d) error("Error, edf_objfun, operator(), bad dimension");

  INTEGER n = cols(sim);
  realmat simulation((lag+1)*d,n-lag);

  for (INTEGER t=lag+1; t<=n; t++) {
    for (INTEGER j=0; j<=lag; ++j) {
      for (INTEGER i=1; i<=d; i++) {
        simulation(j*d+i,t-lag) = sim(i,t-j);
      } 
    }
  }

  libcpp::edf(simulation,data,edf_sim);

  REAL obj = libcpp::edfobj(edf_dat,edf_sim); 

  return obj;
}  

objfun_base* emm::edf_objfun::new_objfun() 
{ 
  objfun_base* r = new(nothrow) edf_objfun(*this); 
  if (r == 0) error("Error, edf_objfun, operator new failed");
  return r; 
}

void emm::edf_objfun::delete_objfun(objfun_base* objfun_ptr)
{ 
  delete objfun_ptr; 
}

bool emm::edf_objfun::write_diagnostics(const char* filename) const
{
  ofstream os(filename);
  if (!os) {
     std::string msg("Error, realmat, write_diagnostics, cannot open ");
     error(msg + filename);
  }

  os << "    t         edf_dat         edf_sim  relative error\n";

  INTEGER t = lag;
  INTEGER r = edf_dat.size();
  for (INTEGER i=1; i<=r; ++i) {
    ++t;
    REAL rel_err = (edf_sim[i] - edf_dat[i])/edf_dat[i];
    os << fmt('d',5,t) << ' '
       << fmt('f',15,5,edf_dat[i]) << ' '
       << fmt('f',15,5,edf_sim[i]) << ' '
       << fmt('f',15,5,rel_err) << '\n';
  }

  os.flush();
  return os.good();
}

